"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeSocketForAck = exports.batchFilter = exports.getIpv4LanIpVerifier = exports.getIpAddressList = exports.realTimeout = exports.runWithTimeout = exports.unZipipData = exports.zipData = exports.parseDomain = exports.genRandomString = void 0;
const logger_1 = require("./logger");
const zlib_1 = require("zlib");
const setting_1 = require("./setting");
const os_1 = require("os");
const chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890';
function genRandomString(length) {
    let res = '';
    while (length--)
        res += chars[Math.floor(Math.random() * chars.length)];
    return res;
}
exports.genRandomString = genRandomString;
/**
 * Parse a `domain:port` string.
 * If there is no `:port` part in given string, the port returned will be 80;
 */
function parseDomain(domain) {
    const domainAndPort = domain.split(':');
    return {
        domain: domainAndPort[0],
        port: domainAndPort.length > 1 ? Number(domainAndPort[1]) : 80,
    };
}
exports.parseDomain = parseDomain;
function zipData(data) {
    return new Promise((r, j) => {
        zlib_1.deflate(data, (err, res) => {
            if (err)
                j(err);
            else
                r(res);
        });
    });
}
exports.zipData = zipData;
function unZipipData(data) {
    return new Promise((r, j) => {
        zlib_1.unzip(data, (err, res) => {
            if (err)
                j(err);
            else
                r(res);
        });
    });
}
exports.unZipipData = unZipipData;
function runWithTimeout(worker, timeout, timeoutValue) {
    if (timeout === Infinity)
        return worker;
    return Promise.race([
        worker,
        new Promise((r) => setTimeout(() => r(timeoutValue), timeout)),
    ]);
}
exports.runWithTimeout = runWithTimeout;
/**
 *
 * The native setTimeout function will ignore time-cost during system sleeping.
 *
 * So use this to gain "realTimeout", which using system time to calculate wether time is out or not.
 *
 * Working by polling and native setTimeout with time unit setting on "Settings.timeoutUnit".
 *
 */
exports.realTimeout = (() => {
    let cbPairs = [];
    let intervalCbPairs = [];
    const tryRun = (cb) => {
        try {
            cb();
        }
        catch (e) {
            logger_1.logger.error(setting_1.ErrorLevel.important, undefined, undefined, 'error in realTimeout cb: ', e);
        }
    };
    const deamonCb = () => {
        setTimeout(deamonCb, setting_1.Settings.timeoutUnit);
        const now = Date.now();
        cbPairs = cbPairs.filter((v) => {
            if (v.at > now)
                return true;
            tryRun(v.cb);
        });
        intervalCbPairs.forEach((v) => {
            if (!v.last || v.last + v.interval < now) {
                tryRun(v.cb);
                v.last = Date.now();
            }
        });
    };
    deamonCb();
    return {
        setTimeout: (cb, timeout) => cbPairs.push({ at: Date.now() + timeout, cb }),
        clearTimeout: (cb) => (cbPairs = cbPairs.filter((v) => v.cb !== cb)),
        setInterval: (cb, interval) => intervalCbPairs.push({ interval, cb }),
        clearInterval: (cb) => (intervalCbPairs = intervalCbPairs.filter((v) => v.cb !== cb)),
    };
})();
const getIpAddressList = () => {
    let ipList = [];
    Object.entries(os_1.networkInterfaces()).forEach(([dev, info]) => {
        if (info)
            ipList = ipList.concat(info.map((v) => v.address));
    });
    return ipList;
};
exports.getIpAddressList = getIpAddressList;
const ipv4ToByteStr = (ipv4) => {
    const numStrs = ipv4.split('.');
    let res = '';
    for (const str of numStrs) {
        const num = Number.parseInt(str);
        if (isNaN(num))
            return '';
        res += num.toString(2).padStart(8, '0');
    }
    return res;
};
const getIpv4LanIpVerifier = () => {
    const maskedIpByteList = [];
    Object.entries(os_1.networkInterfaces()).forEach(([dev, info]) => {
        info === null || info === void 0 ? void 0 : info.forEach((v) => {
            if (v.family === 'IPv4') {
                const length = ipv4ToByteStr(v.netmask).indexOf('0');
                if (length <= 0)
                    return;
                maskedIpByteList.push(ipv4ToByteStr(v.address).slice(0, length));
            }
        });
    });
    return (ip) => {
        const ipBytes = ipv4ToByteStr(ip);
        if (!ipBytes)
            return false;
        return !!maskedIpByteList.find((v) => ipBytes.slice(0, v.length) === v);
    };
};
exports.getIpv4LanIpVerifier = getIpv4LanIpVerifier;
const batchFilter = (datas, cb, batchCount = 30) => __awaiter(void 0, void 0, void 0, function* () {
    if (batchCount <= 0)
        batchCount = 3;
    const res = [];
    const src = [...datas];
    yield Promise.all(new Array(batchCount).fill(1).map(() => __awaiter(void 0, void 0, void 0, function* () {
        while (src.length) {
            const item = src.pop();
            if (item && (yield cb(item)))
                res.push(item);
        }
    })));
    return res;
});
exports.batchFilter = batchFilter;
const writeSocketForAck = (sock, data, timeout = 2000) => new Promise((res, rej) => {
    const onData = (data) => {
        clearTimeout(handler);
        sock.removeListener('data', onData);
        res(data);
    };
    sock.on('data', onData);
    sock.write(data);
    const handler = setTimeout(() => {
        sock.removeListener('data', onData);
        rej('writeSocketForAck timeout');
    }, timeout);
});
exports.writeSocketForAck = writeSocketForAck;
//# sourceMappingURL=util.js.map