"use strict";
/**
 * Protocol processor for Socks5
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProtocolSocks5 = void 0;
const types_1 = require("./../common/types");
const util_1 = require("./../common/util");
const lodash_1 = require("lodash");
const node_fetch_1 = __importDefault(require("node-fetch"));
const NO_AUTH_ACK = Buffer.from([5, 0]);
const NO_AUTH_REQ = Buffer.from([5, 1, 0]);
const parseAddrPort = (data) => {
    let resData, addr;
    // TODO: important check ipv6 and ipv4 addr
    if (data[3] === 1) {
        addr = data.slice(4, 8).join('.');
        resData = data.slice(8);
    }
    else if (data[3] === 3) {
        const end = 5 + data[4];
        addr = String(data.slice(5, end));
        // TODO: check chinese char.
        resData = data.slice(end);
    }
    else if (data[3] === 4) {
        addr = lodash_1.chunk(Array.from(data.slice(4, 20)).map((v) => v.toString(16).padStart(2, '0')), 2)
            .map((v) => v.join(''))
            .join(':');
        // TODO: check ipv6.
        resData = data.slice(20);
    }
    else {
        throw new types_1.ErrorProtocolProcessing('Parse socks5 addr failed');
    }
    if (resData.length !== 2)
        throw new types_1.ErrorProtocolProcessing('Parse socks5 port failed: illegal data length');
    return { addr, port: resData.readUInt16BE() };
};
/**
 * Minimal socks5 protocol implemention.
 *
 * TODO: Authority method implemention.
 */
class ProtocolSocks5 extends types_1.ProtocolBase {
    constructor() {
        super(...arguments);
        this.recvDatas = [];
        this.protocol = 'socks5';
        this.targetPackage = null;
    }
    process(data) {
        return __awaiter(this, void 0, void 0, function* () {
            if (data[0] === 5 && data.length === 2 + data[1]) {
                this.recvDatas.push(data);
                const recv = yield util_1.writeSocketForAck(this.sock, NO_AUTH_ACK).catch((e) => false);
                if (!recv || recv[0] !== 5)
                    throw new types_1.ErrorProtocolProcessing('Socks5 communication error');
                if (recv[1] !== 1)
                    throw new types_1.ErrorProtocolProcessing('Unsupported socks5 method');
                Object.assign(this, parseAddrPort(recv));
                this.recvDatas.push(recv);
                this.targetPackage = recv;
                return this;
            }
        });
    }
    doFeedBack(code) {
        if (this.targetPackage) {
            const ack = Buffer.from(this.targetPackage);
            ack[1] = code;
            this.sock.write(ack);
        }
    }
    /** TODO: dose socks5 need fail feedback? */
    doFailFeedback() {
        return __awaiter(this, void 0, void 0, function* () {
            this.doFeedBack(4);
        });
    }
    doConnectedFeedback() {
        return __awaiter(this, void 0, void 0, function* () {
            this.doFeedBack(0);
        });
    }
    takeOver(targets) {
        const destSock = this.connectFunc(targets, this);
        destSock.on('data', (data) => this.sock.write(data));
        destSock.on('end', () => this.sock.destroy());
        destSock.on('error', (err) => {
            this.sock.destroy();
        });
        destSock.on('connect', () => this.doConnectedFeedback());
        this.sock.on('data', (data) => {
            destSock.write(data);
        });
        this.sock.on('end', destSock.destroy);
        this.sock.on('error', (err) => {
            destSock.destroy();
        });
    }
    onLogicConnect(target, targetSock) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!target.notProxy) {
                let ack = yield util_1.writeSocketForAck(targetSock, NO_AUTH_REQ, 1500);
                if (!ack.equals(NO_AUTH_ACK))
                    throw new types_1.ErrorProtocolProcessing('Socks5 error: unknown ack from next proxy');
                if (!this.targetPackage)
                    throw new types_1.ErrorProtocolProcessing('Socks5 error: lack targetPackage');
                ack = yield util_1.writeSocketForAck(targetSock, this.targetPackage, 2500);
                if (ack[0] !== 5 || ack[1] !== 0 || !ack.slice(2).equals(this.targetPackage.slice(2))) {
                    throw new types_1.ErrorProtocolProcessing('Socks5 error: wrong ack message');
                }
                return 2;
            }
            return 0;
        });
    }
    doIdleTimeoutVerify(target) {
        return __awaiter(this, void 0, void 0, function* () {
            // TODO: verify through next proxy if necessary.
            if (!target.notProxy)
                return true;
            return yield node_fetch_1.default(`${this.port === 443 ? 'https' : 'http'}://${this.addr}${[80, 443].includes(this.port) ? '' : this.port}`, {
                method: 'GET',
                redirect: 'manual',
                timeout: 2000,
            })
                .then((res) => !!res)
                .catch(() => false);
        });
    }
}
exports.ProtocolSocks5 = ProtocolSocks5;
//# sourceMappingURL=socks5.js.map