"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isLanIpv4 = exports.pingDomain = exports.resolveDomainIps = exports.getTargets = exports.diagnoseDomain = exports.judgeToSaveCache = exports.trySaveCache = exports.tryRestoreCache = exports.domainStatDescMap = exports.notExactlyGoodStats = exports.domainStatsMap = void 0;
const logger_1 = require("./../common/logger");
const setting_1 = require("./../common/setting");
const net_1 = __importDefault(require("net"));
const dns_1 = require("dns");
const ping_1 = __importDefault(require("ping"));
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
const types_1 = require("../common/types");
const util_1 = require("../common/util");
const lodash_1 = require("lodash");
const is_online_1 = __importDefault(require("is-online"));
/**
 * Map to store diagnostic infos.
 *
 * DomainChannelStats is sorted by healthy
 *
 * Key is `domain:port` string.
 */
exports.domainStatsMap = new Map();
exports.notExactlyGoodStats = (() => {
    const notGoodDomainMap = new Map();
    return {
        countNotGood(domain) {
            return notGoodDomainMap.get(domain) || 0;
        },
        feedback(notGood, domain) {
            if (exports.isLanIpv4(domain))
                return;
            if (!notGood && !notGoodDomainMap.has(domain))
                return;
            let stats = notGoodDomainMap.get(domain) || 0;
            if (notGood)
                stats++;
            else
                stats -= 0.25;
            if (stats < 0)
                notGoodDomainMap.delete(domain);
            else
                notGoodDomainMap.set(domain, stats);
        },
        /**
         * Clear notGood count of one domain or all.
         */
        clear(domain) {
            if (domain)
                notGoodDomainMap.delete(domain);
            else
                notGoodDomainMap.clear();
        },
    };
})();
/**
 * Map to count domain request times and last active time;
 *
 * Key is `domain:port` string.
 */
exports.domainStatDescMap = new Map();
const judgeDomainActivity = (desc) => {
    const now = Math.floor(Date.now() / 3600000);
    if (!desc.at)
        desc.at = now;
    if (desc.at + setting_1.Settings.cacheDomainLife > now)
        return true;
    desc.at = now;
    desc.count = Math.floor(desc.count / 3);
    return !!desc.count;
};
const tryRestoreCache = () => __awaiter(void 0, void 0, void 0, function* () {
    const cacheFile = path_1.default.resolve(setting_1.Settings.cacheFile);
    if (fs_1.default.existsSync(cacheFile)) {
        try {
            const data = JSON.parse(String(yield util_1.unZipipData(fs_1.default.readFileSync(cacheFile))));
            if (data.domainReqCountsDec) {
                data.domainReqCountsDec = data.domainReqCountsDec.filter(judgeDomainActivity);
                exports.domainStatDescMap = new Map(data.domainReqCountsDec.map((v) => [v.dAndP, v]));
                // Ping test after start
                const dataToPing = Array.from(data.domainReqCountsDec);
                const pingTest = () => __awaiter(void 0, void 0, void 0, function* () {
                    const toPing = dataToPing
                        .splice(0, setting_1.Settings.pingBatchCount)
                        .filter((d) => !exports.domainStatsMap.has(d.dAndP));
                    yield Promise.all(toPing.map((d) => {
                        const dAndP = util_1.parseDomain(d.dAndP);
                        return util_1.runWithTimeout(exports.diagnoseDomain(dAndP.domain, dAndP.port, false, true, true), 3000, undefined);
                    }));
                    if (!dataToPing.length)
                        return;
                    else
                        process.nextTick(pingTest);
                });
                process.nextTick(pingTest);
            }
        }
        catch (e) {
            logger_1.logger.error(setting_1.ErrorLevel.dangerous, undefined, undefined, 'Parse cache data failed: ' + e.message);
        }
    }
});
exports.tryRestoreCache = tryRestoreCache;
const trySaveCache = () => __awaiter(void 0, void 0, void 0, function* () {
    const cacheData = {};
    cacheData.domainReqCountsDec = Array.from(exports.domainStatDescMap.values()).sort((a, b) => b.count - a.count);
    try {
        fs_1.default.writeFileSync(setting_1.Settings.cacheFile, yield util_1.zipData(Buffer.from(JSON.stringify(cacheData))));
    }
    catch (e) {
        console.error('Save cache data failed: ' + e.message);
    }
});
exports.trySaveCache = trySaveCache;
exports.judgeToSaveCache = (() => {
    let lastUpdatedAt = 0;
    let lastDomainCount = 0;
    return () => __awaiter(void 0, void 0, void 0, function* () {
        if (exports.domainStatDescMap.size === lastDomainCount)
            return;
        if (lastUpdatedAt &&
            exports.domainStatDescMap.size - lastDomainCount < 3 &&
            Date.now() - lastUpdatedAt < 10 * 60 * 1000)
            return;
        yield exports.trySaveCache();
        lastUpdatedAt = Date.now();
        lastDomainCount = exports.domainStatDescMap.size;
    });
})();
const judgeToSaveCacheDebounced = lodash_1.debounce(exports.judgeToSaveCache, 2000);
/**
 * Use ping, tcp-ping tools to test targets and feedBack targets
 * If all target tested, return best two target.
 * Test will be skipped if no proxy defined in Setting.proxys.
 * TODO: auto rediagnose
 * TODO: spread test on next tunnel-proxy node
 */
const diagnoseDomain = (domain, port, rediagnose = false, forceSync = false, ignoreCount = false) => __awaiter(void 0, void 0, void 0, function* () {
    const dAndP = `${domain}:${port}`;
    if (!ignoreCount) {
        const desc = exports.domainStatDescMap.get(dAndP) || { dAndP, count: 0 };
        desc.count++;
        desc.at = Math.floor(Date.now() / 3600000);
        if (desc.count === 1) {
            exports.domainStatDescMap.set(dAndP, desc);
            judgeToSaveCacheDebounced();
        }
    }
    let statsList = rediagnose ? [] : exports.domainStatsMap.get(dAndP) || [];
    if (!statsList.length || !(yield verifyTtl(statsList))) {
        setting_1.Settings.proxys.forEach((target) => statsList.push(new types_1.DomainChannelStats(domain, port, dAndP, target)));
        yield new Promise((r) => __awaiter(void 0, void 0, void 0, function* () {
            const mStatsList = [...statsList];
            const ips = yield exports.resolveDomainIps(domain);
            if (ips.length) {
                if (!setting_1.Settings.proxys.length || (!forceSync && setting_1.Settings.pingAsync)) {
                    statsList.push(new types_1.DomainChannelStats(domain, port, dAndP, { ip: ips[0].address, port, notProxy: true }, ips[0].ttl));
                    r(undefined);
                    if (!setting_1.Settings.proxys.length)
                        return;
                }
                const pingResList = yield Promise.all(ips.map((ip) => exports.pingDomain(ip.address)));
                const localStatsis = [];
                pingResList.forEach((res, i) => {
                    const lossPct = 
                    // @ts-ignore
                    res.packetLoss === 'unknown' ? Infinity : Number(res.packetLoss);
                    if (lossPct <= setting_1.Settings.maxPkgLossPct) {
                        const stats = new types_1.DomainChannelStats(domain, port, dAndP, {
                            ip: ips[i].address,
                            port,
                            notProxy: true,
                        }, ips[i].ttl);
                        stats.latency = res.time === 'unknown' ? Infinity : res.time;
                        stats.pkgLostPct = lossPct;
                        stats.status =
                            lossPct > 0 ||
                                exports.notExactlyGoodStats.countNotGood(domain) >
                                    setting_1.Settings.notExactlyGoodCountLimit
                                ? 'bad'
                                : stats.latency < setting_1.Settings.maxGoodLatency &&
                                    !exports.notExactlyGoodStats.countNotGood(domain)
                                    ? 'good'
                                    : 'work';
                        localStatsis.push(stats);
                    }
                });
                if (localStatsis.length) {
                    localStatsis
                        .sort((a, b) => a.pkgLostPct - b.pkgLostPct)
                        .slice(0, 2)
                        .forEach((s) => {
                        if (s.status === 'good')
                            mStatsList.unshift(s);
                        else
                            mStatsList.push(s);
                    });
                }
                if (!forceSync && setting_1.Settings.pingAsync) {
                    if (mStatsList.length)
                        exports.domainStatsMap.set(dAndP, mStatsList);
                }
                else {
                    statsList = mStatsList;
                    r(undefined);
                }
            }
        }));
    }
    if (!statsList.length) {
        const oldStats = exports.domainStatsMap.get(dAndP);
        if (!(oldStats === null || oldStats === void 0 ? void 0 : oldStats.length))
            return [];
        else
            statsList = oldStats;
    }
    exports.domainStatsMap.set(dAndP, statsList);
    const notGoodCount = exports.notExactlyGoodStats.countNotGood(domain);
    // pick stats
    if (statsList[0].status === 'good' && !notGoodCount)
        return [statsList[0]];
    const filteredStats = notGoodCount < setting_1.Settings.notExactlyGoodCountLimit
        ? statsList.filter((s) => s.status !== 'bad')
        : statsList.filter((s) => !s.target.notProxy);
    const targets = filteredStats.length ? filteredStats : statsList;
    if (targets.length)
        return targets;
    return statsList;
});
exports.diagnoseDomain = diagnoseDomain;
const getTargets = (addr, port) => __awaiter(void 0, void 0, void 0, function* () {
    const target = setting_1.getDomainProxy(addr);
    if (target)
        return [new types_1.DomainChannelStats(addr, port, `${addr}:${port}`, target)];
    return yield exports.diagnoseDomain(addr, port);
});
exports.getTargets = getTargets;
// TODO:
//export const targetFeedBack = (target: Target, latency: number) => {};
const stuckResolveDomainMap = new Map();
// TODO: some domain's dns resolving  cost more than tens of seconds
const resolveDomainIps = (domain) => __awaiter(void 0, void 0, void 0, function* () {
    if (net_1.default.isIP(domain))
        return [{ address: domain, ttl: Infinity }];
    const res = yield util_1.runWithTimeout(new Promise((resolve) => {
        const stucks = stuckResolveDomainMap.get(domain) || [];
        stucks.push(resolve);
        if (stucks.length === 1)
            stuckResolveDomainMap.set(domain, stucks);
        else
            return;
        const r = (res) => {
            (stuckResolveDomainMap.get(domain) || []).forEach((resolve) => resolve(res));
            stuckResolveDomainMap.delete(domain);
        };
        const cb = (err, ips) => {
            count--;
            if (err || !ips.length) {
                if (!count)
                    r([]);
            }
            else
                r(ips);
        };
        let count = [
            dns_1.resolve4(domain, { ttl: true }, cb),
            ...(setting_1.Settings.useIpv6 ? [dns_1.resolve6(domain, { ttl: true }, cb)] : []),
        ].length;
    }), setting_1.Settings.proxys.length ? setting_1.Settings.dnsTimeout : Infinity, []);
    return res;
});
exports.resolveDomainIps = resolveDomainIps;
const failedPingRes = {
    host: '',
    alive: false,
    output: '',
    time: Infinity,
    times: [],
    min: 'Infinity',
    max: 'Infinity',
    avg: 'Infinity',
    stddev: '0',
    packetLoss: '100',
    numeric_host: '',
};
const pingDomain = (domain, count = 10) => {
    const failedRes = Object.assign(Object.assign({}, failedPingRes), { host: domain });
    return util_1.runWithTimeout(new Promise((r) => {
        ping_1.default.promise
            .probe(domain, {
            v6: domain.includes(':'),
            timeout: setting_1.Settings.pingTimeout,
            extra: ['-c', '' + count, '-i', '0.2'],
        })
            .then(r)
            .catch((e) => {
            logger_1.logger.error(setting_1.ErrorLevel.notice, undefined, undefined, 'Error: ping failed:\n\t', domain, e);
            r(failedRes);
        });
    }), count * 1000, failedRes);
};
exports.pingDomain = pingDomain;
/**
 * Verify dns ttl.
 * @param reuseThesameIp true to return true if ttl expired but new dns resolve result contains old ip.
 * @param forceExpired true to force verify even ttl remains;
 */
const verifyTtl = (stats, margin = 0, reuseThesameIp = true, forceExpired = false) => __awaiter(void 0, void 0, void 0, function* () {
    stats = stats.filter((v) => v.target.notProxy);
    const now = Date.now();
    const expireds = forceExpired
        ? stats
        : stats.filter((v) => v.updateAtMili + v.ttl * 1000 < now + margin);
    if (!expireds.length)
        return true;
    if (!reuseThesameIp)
        return false;
    const res = yield new Promise((r) => {
        let count = expireds.length;
        expireds.forEach((st) => __awaiter(void 0, void 0, void 0, function* () {
            const res = yield exports.resolveDomainIps(st.domain);
            if (count) {
                count--;
                if (!res.find((ip) => ip.address === st.target.ip)) {
                    count = 0;
                    r(false);
                }
                else {
                    st.updateAtMili = Date.now();
                    if (count <= 0) {
                        r(true);
                    }
                }
            }
        }));
    });
    return res;
});
exports.isLanIpv4 = util_1.getIpv4LanIpVerifier();
// TODO: find better way to refresh
(() => {
    const waitMilli = 30000;
    const { setTimeout } = util_1.realTimeout;
    let lock = '';
    const cycleRefreshTtl = (ignoreLock = false, forceUpdate = false) => __awaiter(void 0, void 0, void 0, function* () {
        if (lock && !ignoreLock)
            return;
        if (!(yield is_online_1.default())) {
            setTimeout(cycleRefreshTtl, 5000);
            return;
        }
        const mLock = Math.random().toString();
        lock = mLock;
        logger_1.logger.log(setting_1.LogLevel.detail, undefined, undefined, 'Domain ttl refresh start');
        if (ignoreLock)
            util_1.realTimeout.clearTimeout(cycleRefreshTtl);
        const start = Date.now();
        const existStats = Array.from(exports.domainStatsMap.values());
        const stats = forceUpdate
            ? existStats
            : yield util_1.batchFilter(existStats, forceUpdate ? () => __awaiter(void 0, void 0, void 0, function* () { return true; }) : (v) => __awaiter(void 0, void 0, void 0, function* () { return !(yield verifyTtl(v, waitMilli)); }));
        const worker = () => __awaiter(void 0, void 0, void 0, function* () {
            if (stats.length) {
                yield Promise.all(stats
                    .splice(0, Math.min(setting_1.Settings.pingBatchCount, 20))
                    .map((v) => util_1.runWithTimeout(exports.diagnoseDomain(v[0].domain, v[0].port, true, true, true), 3000, undefined)));
                if (lock !== mLock)
                    return;
                process.nextTick(worker);
            }
            else {
                logger_1.logger.log(setting_1.LogLevel.detail, undefined, undefined, 'Domain ttl refresh done');
                setTimeout(cycleRefreshTtl, waitMilli - (Date.now() - start));
                lock = '';
            }
        });
        worker();
    });
    cycleRefreshTtl();
    // listening for ip change
    let lastIpList = null;
    util_1.realTimeout.setInterval(() => {
        const ipList = util_1.getIpAddressList();
        if (lastIpList) {
            if (JSON.stringify(lastIpList) !== JSON.stringify(ipList)) {
                logger_1.logger.log(setting_1.LogLevel.notice, undefined, undefined, 'System Ip list change');
                exports.notExactlyGoodStats.clear();
                cycleRefreshTtl(true, true);
                exports.isLanIpv4 = util_1.getIpv4LanIpVerifier();
            }
        }
        lastIpList = ipList;
    }, 1000);
    // Set interval to judge cached domain life.
    util_1.realTimeout.setInterval(() => {
        Array.from(exports.domainStatDescMap.entries()).forEach(([dAndP, desc]) => {
            if (!judgeDomainActivity(desc)) {
                exports.domainStatDescMap.delete(dAndP);
                exports.domainStatsMap.delete(dAndP);
            }
        });
        exports.judgeToSaveCache();
    }, 6000 || 3600000 * 12);
})();
//# sourceMappingURL=channelDiagnostic.js.map